'''
Base code for 'optitrack' feature, compatible with Optitrack motiontracker
'''
import threading
import time
import numpy as np
from ..source import DataSourceSystem
import socket
import re
import aopy

UDP_PORT = 11999

class System(DataSourceSystem):
    '''
    Optitrack DataSourceSystem collects motion tracking data via UDP packets using natnet depacketizer
    '''
    update_freq = 240 # This may not always be the case, but lower frequencies will still work, just waste space in the circular buffer
    dtype = np.dtype(('float', (6,)))

    def start(self):
        '''
        Just set the callback function
        '''
        # Format is "frame_count&le_x&le_y&le_diam&re_x&re_y&re_diam"

        self.sock = socket.socket(socket.AF_INET, # Internet
                            socket.SOCK_DGRAM) # UDP
        self.sock.bind(("", UDP_PORT))   

    def stop(self):
        self.sock.close()
    
    def get(self):
        '''
        Main logic -- parse the motion tracking data into a defined datatype
        '''      
        data, addr = self.sock.recvfrom(1024) # buffer size is 1024 bytes
        items = data.decode('utf-8').split('&')
        frame_count = float(items[0]) # ignored

        # Careful with the x and y positions as they can be "Inf"
        try:
            le_x = float(items[1])
        except:
            le_x = np.nan
        try:
            le_y = float(items[2])
        except:
            le_y = np.nan
        try:
            re_x = float(items[4])
        except:
            re_x = np.nan
        try:
            re_y = float(items[5])
        except:
            re_y = np.nan
        
        le_diam = float(items[3])
        re_diam = float(items[6])

        # Pack into (1,6) array
        coords = np.array([le_x, le_y, re_x, re_y, le_diam, re_diam])
        coords = np.expand_dims(coords, axis=0)
        return coords


#################
# Simulated data
#################
class SimulatedEye(threading.Thread):
    '''
    This is not tested
    '''
    update_rate = 240

    def __init__(self, radius=(0.2,0.04,0.21,0.05,20,21), speed=(0.5,1,0.4,1.2,0.1, 0.1)):
        self.stime = time.time()
        self.radius = radius
        self.speed = speed
        self.frame_count = 0
        self.sock = socket.socket(socket.AF_INET, # Internet
                            socket.SOCK_DGRAM) # UDP

    def run_once(self):
        '''
        Fake some motion data
        '''
        time.sleep(1./240)
        ts = (time.time() - self.stime)
        coords = np.multiply(self.radius, np.cos(np.divide(ts, self.speed) * 2 * np.pi))
        self.send_data(coords)

    def send_data(self, coords):
        data = f"{self.frame_count}&{coords[0]}&{coords[1]}&{coords[2]}&{coords[3]}&{coords[4]}&{coords[5]}"
        self.sock.sendto(data.encode(), ("localhost", UDP_PORT))
        self.frame_count += 1

    def run(self):
        while True:
            time.sleep(1./self.update_rate)
            self.run_once()

########################
# Playback from csv file
########################

class PlaybackEye(SimulatedEye):
    '''
    This is not tested
    '''
    def __init__(self, data_dir, filename):
        self.data, metadata = aopy.data.load_bmi3d_hdf_table(data_dir, filename, 'task')

    def run_once(self):
        '''
        Fake some motion data
        '''
        coords = self.data['eyedata'][self.frame_count]
        self.send_data(coords)







from .target_capture_task import ScreenTargetCapture
from riglib.experiment import traits
import os
from riglib.audio import AudioPlayer

audio_path = os.path.join(os.path.dirname(__file__), '../riglib/audio')

class ScreenTargetCapture_ReadySet(ScreenTargetCapture):

    '''
    Center out task with ready set go auditory cues. Cues separated by 500 ms and participant is expected to move on final go cue. Additionally, participant must move out
    of center circle (mustmv_time) parameter or there will be an error. 
    '''
    
    status = dict(
        wait = dict(start_trial="target"),
        target = dict(enter_target="hold", timeout="timeout_penalty"),
        hold = dict(leave_target="hold_penalty", hold_complete_center="prepbuff", hold_complete_periph='reward'),
        prepbuff = dict(leave_target="hold_penalty", prepbuff_complete="delay"),
        delay = dict(leave_target="delay_penalty", delay_complete="leave_center"),
        leave_center = dict(leave_target="targ_transition", mustmv_complete="hold_penalty"),
        targ_transition = dict(trial_complete="reward", trial_abort="wait", trial_incomplete="target"),
        timeout_penalty = dict(timeout_penalty_end="targ_transition", end_state=True),
        hold_penalty = dict(hold_penalty_end="targ_transition", end_state=True),
        delay_penalty = dict(delay_penalty_end="targ_transition", end_state=True),
        reward = dict(reward_end="wait", stoppable=False, end_state=True)
    )

    #exclude_parent_traits = ['delay_time']

    prepbuff_time = traits.Float(.2, desc="How long after acquiring center target before peripheral target appears")
    mustmv_time = traits.Float(.2, desc="Must leave center target within this time after auditory go cue.")
    
    files = [f for f in os.listdir(audio_path) if '.wav' in f]
    ready_set_sound = traits.OptionsList(files, desc="File in riglib/audio to play on each reward")


    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.ready_set_player = AudioPlayer(self.ready_set_sound)
        # sound_time = self.ready_set_player.get_length()
        # self.delay_time = 
        # self.prepbuff_time = (sound_time + mustmv_time) - self.delay_time

    ###Test Functions ###
    def _test_hold_complete_center(self, time_in_state):
        '''
        Test whether the center target is held long enough to declare the
        trial a success 

        Possible options
            - Target held for the minimum requred time (implemented here)
            - Sensorized object moved by a certain amount
            - Sensorized object moved to the required location
            - Manually triggered by experimenter
        '''
        return self.target_index == 0 and time_in_state > self.hold_time
    
    def _test_hold_complete_periph(self, time_in_state):
        '''
        Test whether the peripheral target is held long enough to declare the
        trial a success 

        Possible options
            - Target held for the minimum requred time (implemented here)
            - Sensorized object moved by a certain amount
            - Sensorized object moved to the required location
            - Manually triggered by experimenter
        '''
        return self.target_index == 1 and time_in_state > self.hold_time
    
    def _test_prepbuff_complete(self, time_in_state):
        '''
        Test whether the target is held long enough to declare the
        trial a success

        Possible options
            - Target held for the minimum requred time (implemented here)
            - Sensorized object moved by a certain amount
            - Sensorized object moved to the required location
            - Manually triggered by experimenter
        '''
        return time_in_state > self.prepbuff_time
    
    def _test_mustmv_complete(self, time_in_state):
        '''
        Test whether the target is exited in time. Return of true for mustmv sends to penalty state.  

        Possible options
            - Target left before the minimum requred time (implemented here)
            - Sensorized object moved by a certain amount
            - Sensorized object moved to the required location
            - Manually triggered by experimenter
        '''
        return time_in_state > self.mustmv_time

    ### State Functions ###
    def _start_prepbuff(self):

        #self.sync_event('CURSOR_LEAVE_TARGET', self.gen_indices[self.target_index])
        self.ready_set_player.play()

    def _start_leave_center(self):

        if self.target_index == 0:
            #self.targets[0].hide()
            self.sync_event('CENTER_TARGET_OFF', self.gen_indices[self.target_index])

    def _start_hold_penalty(self):
        if hasattr(super(), '_start_hold_penalty'):
            super()._start_hold_penalty()
        self.ready_set_player.stop()
    
    def _start_delay_penalty(self):
        if hasattr(super(), '_start_delay_penalty'):
            super()._start_delay_penalty()
        self.ready_set_player.stop()


"""Finite state machine implementation """
import time
import threading
import traceback
import io

from collections import OrderedDict

min_per_hour = 60.
sec_per_min = 60.


class FSMTable(object):
    def __init__(self, **kwargs):
        self.states = OrderedDict()
        for state_name, transitions in list(kwargs.items()):
            self.states[state_name] = transitions

    def __getitem__(self, key):
        return self.states[key]

    def get_possible_state_transitions(self, current_state):
        return list(self.states[current_state].items())

    def _lookup_next_state(self, current_state, transition_event):
        return self.states[current_state][transition_event]

    def __iter__(self):
        return list(self.states.keys()).__iter__()

    @property
    def trial_end_states(self):
        return [state for state, transitions in self.states.items() if transitions.end_state]

    @staticmethod
    def construct_from_dict(status):
        outward_transitions = OrderedDict()
        for state in status:
            outward_transitions[state] = StateTransitions(**status[state])
        return FSMTable(**outward_transitions)


class StateTransitions(object):
    def __init__(self, stoppable=True, end_state=False, **kwargs):
        self.state_transitions = OrderedDict()
        for event, next_state in list(kwargs.items()):
            self.state_transitions[event] = next_state

        if stoppable and not ('stop' in self.state_transitions):
            self.state_transitions['stop'] = None
        self.end_state = end_state

    def __getitem__(self, key):
        return self.state_transitions[key]

    def __iter__(self):
        transition_events = list(self.state_transitions.keys())
        return transition_events.__iter__()

    def items(self):
        return list(self.state_transitions.items())


class Clock(object):
    def tick(self, fps):
        import time
        time.sleep(1.0/fps)
    def get_time(self):
        return 0

class FSM(object):
    status = FSMTable(
        wait = StateTransitions(start_trial="trial", premature="penalty", stop=None),
        trial = StateTransitions(correct="reward", incorrect="penalty", timeout="penalty"),
        reward = StateTransitions(post_reward="wait"),
        penalty = StateTransitions(post_penalty="wait"),
    )
    state = "wait"
    debug = False
    fps = 60 # frames per second

    log_exclude = set()  # List out state/trigger pairs to exclude from logging

    def __init__(self, *args, **kwargs):
        self.verbose = kwargs.pop('verbose', False)

        # state and event transitions
        self.state_log = []
        self.event_log = []

        self.clock = Clock()

        # Timestamp for rough loop timing
        self.last_time = self.get_time()
        self.cycle_count = 0        

    @property 
    def update_rate(self):
        '''
        Attribute for update rate of task. Using @property in case any future modifications
        decide to change fps on initialization
        '''
        return 1./self.fps        

    def print_to_terminal(self, *args):
        '''
        Print to the terminal rather than the websocket if the websocket is being used by the 'Notify' feature (see db.tasktrack)
        '''
        if len(args) == 1:
            print(args[0])
        else:
            print(args)        

    def init(self):
        '''Interface for child classes to run initialization code after object
        construction'''
        pass

    def run(self):
        '''
        Generic method to run the finite state machine of the task. Code that needs to execute 
        imediately before the task starts running in child classes should be of the form:

        def run(self):
            do stuff
            try:
                super(class_name, self).run()
            finally:
                clean up stuff

        The try block may or may not be necessary. For example, if you're opening a UDP port, you may want to always
        close the socket whether or not the main loop executes properly so that you don't loose the 
        reference to the socket. 
        '''

        ## Initialize the FSM before the loop
        self.set_state(self.state)
        
        while self.state is not None:
            if self.debug: 
                # allow ungraceful termination if in debugging mode so that pdb 
                # can catch the exception in the appropriate place
                self.fsm_tick()
            else:
                # in "production" mode (not debugging), try to capture & log errors gracefully
                try:
                    self.fsm_tick()
                except:
                    self.print_to_terminal("Error in FSM tick")
                    self.state = None
                    self.terminated_in_error = True

                    self.termination_err = io.StringIO()
                    traceback.print_exc(None, self.termination_err)
                    self.termination_err.seek(0)

                    self.print_to_terminal(self.termination_err.read())
                    self.termination_err.seek(0)
        if self.verbose: print("end of FSM.run, task state is", self.state)

    def run_sync(self):
        self.init()
        self.run()

    ###########################################################
    ##### Finite state machine (FSM) transition functions #####
    ###########################################################
    def fsm_tick(self):
        '''
        Execute the commands corresponding to a single tick of the event loop
        '''
    
        # Execute commands
        self.exec_state_specific_actions(self.state)

        current_state = self.state

        # iterate over the possible events which could move the task out of the current state
        for event in self.status[current_state]:
            if self.test_state_transition_event(event): # if the event has occurred
                # execute commands to end the current state
                self.end_state(current_state)

                # trigger the transition for the event
                self.trigger_event(event)

                # stop searching for transition events (transition events must be 
                # mutually exclusive for this FSM to function properly)
                break

        # Execute the commands which must run every loop, independent of the FSM state
        # (e.g., running the BMI decoder, updating the display)
        if self.state is not None:
            self._cycle()    


    def test_state_transition_event(self, event):
        event_test_fn_name = "_test_%s" % event
        if hasattr(self, event_test_fn_name):
            event_test_fn = getattr(self, event_test_fn_name)
            time_since_state_started = self.get_time() - self.start_time
            return event_test_fn(time_since_state_started)
        else:
            return False

    def end_state(self, state):
        end_state_fn_name = "_end_%s" % state
        if hasattr(self, end_state_fn_name):
            end_state_fn = getattr(self, end_state_fn_name)
            end_state_fn()

    def start_state(self, state):
        state_start_fn_name = "_start_%s" % state
        if hasattr(self, state_start_fn_name):
            state_start_fn = getattr(self, state_start_fn_name)
            state_start_fn()

    def exec_state_specific_actions(self, state):
        if hasattr(self, "_while_%s" % state):
            getattr(self, "_while_%s" % state)()

    def trigger_event(self, event):
        '''
        Transition the task state to a new state, where the next state depends on the current state as well as the trigger event

        Parameters
        ----------
        event: string
            Based on the current state, a particular event will trigger a particular state transition (Mealy machine)

        Returns
        -------
        None
        '''
        log = (self.state, event) not in self.log_exclude
        if log:  
            self.event_log.append((self.state, event, self.get_time()))

        fsm_edges = self.status[self.state]
        next_state = fsm_edges[event]            
        self.set_state(next_state, log=log)        
        
    def set_state(self, condition, log=True):
        '''
        Change the state of the task

        Parameters
        ----------
        condition: string
            Name of new state. The state name must be a key in the 'status' dictionary attribute of the task

        Returns
        -------
        None
        '''
        # Record the time at which the new state is entered. Used for timed states, e.g., the reward state
        self.start_time = self.get_time()

        if log:
            self.state_log.append((condition, self.start_time))        
        self.state = condition

        self.start_state(condition)

    def get_time(self):
        '''
        Abstraction to get the current time. By default, state transitions are based on wall clock time, not on iteration count.
        To get simulations to run faster than real time, this function must be overwritten.

        Returns
        -------
        float: The current time in seconds
        '''
        return time.time()

    def _cycle(self):
        '''
        Code that needs to run every task loop iteration goes here
        '''
        self.cycle_count += 1
        if self.fps > 0:
            self.clock.tick(self.fps)   

    def iter_time(self):
        '''
        Determine the time elapsed since the last time this function was called
        '''
        start_time = self.get_time()
        loop_time = start_time - self.last_time
        self.last_time = start_time
        return loop_time

    @classmethod
    def parse_fsm(cls):
        '''
        Print out the FSM of the task in a semi-readable form
        '''
        for state in cls.status:
            print('When in state "%s"' % state) 
            for trigger_event, next_state in list(cls.status[state].items()):
                print('\tevent "%s" moves the task to state "%s"' % (trigger_event, next_state))

    @classmethod
    def auto_gen_fsm_functions(cls):
        '''
        Parse the FSM to write all the _start, _end, _while, and _test functions
        '''
        events_to_test = []
        for state in cls.status:
            # make _start function 
            print('''def _start_%s(self): pass''' % state)

            # make _while function
            print('''def _while_%s(self): pass''' % state)
            # make _end function
            print('''def _end_%s(self): pass''' % state)
            for event, _ in cls.status.get_possible_state_transitions(state):
                events_to_test.append(event)

        print("################## State trnasition test functions ##################")

        for event in events_to_test:
            if event == 'stop': continue
            print('''def _test_%s(self, time_in_state): return False''' % event)

    def end_task(self):
        '''
        End the FSM gracefully on the next iteration by setting the task's "stop" flag.
        '''
        self.stop = True      

    def _test_stop(self, ts):
        ''' 
        FSM 'test' function. Returns the 'stop' attribute of the task
        '''
        return self.stop


class ThreadedFSM(FSM, threading.Thread):
    """ FSM + infrastructure to run FSM in its own thread """
    def __init__(self):
        FSM.__init__(self)
        threading.Thread.__init__(self)

    def start(self):
        '''
        From the python docs on threading.Thread:
            Once a thread object is created, its activity must be started by 
            calling the thread's start() method. This invokes the run() method in a 
            separate thread of control.

        Prior to the thread's start method being called, the secondary init function (self.init) is executed.
        After the threading.Thread.start is executed, the 'run' method is executed automatically in a separate thread.

        Returns
        -------
        None
        '''
        self.init()
        threading.Thread.start(self)

    def join(self):
        '''
        Code to run before re-joining the FSM thread 
        '''
        threading.Thread.join(self)



'''
Generic data sink. Sinks run in separate processes and interact with the main process through code here
'''
from .mp_proxy import FuncProxy, RPCProcess
from . import singleton


class DataSink(RPCProcess):
    '''Generic single-channel data sink'''
    def loop_task(self):
        if self.data_pipe.poll(0.001):
            system, data = self.data_pipe.recv()
            self.target.send(system, data)

    def target_destr(self, ret_status, msg):
        self.target.close()
        print("ended datasink")

    def send(self, system, data):
        '''
        Send data to the sink system running in the remote process

        Parameters
        ----------
        system : string
            Name of system (source) from which the data originated
        data : object
            Arbitrary data. The remote sink should know how to handle the data

        Returns
        -------
        None
        '''
        if self.status.value > 0:
            self.data_proxy.pipe.send((system, data))


class SinkManager(singleton.Singleton):
    __instance = None
    ''' Data Sink manager singleton to be used by features '''
    def __init__(self):
        '''Don't call this constructor directly. Use SinkManager.get_instance()'''
        super().__init__()
        self.reset()

    def reset(self):
        self.sinks = []
        self.sources = []
        self.registrations = dict()

    def start(self, output, log_filename='', **kwargs):
        '''
        Create a new sink and register with it all the known sources.

        Parameters
        ----------
        output : type
            Data sink target class
        kwargs : optional kwargs
            arguments passed to the data sink target

        Returns
        -------
        sink : DataSink
            Newly-created data sink
        '''
        # print(("sinkmanager start %s"%output))
        sink = DataSink(target_class=output, target_kwargs=kwargs, log_filename=log_filename)
        sink.start()
        self.registrations[sink] = set()
        for source, dtype in self.sources:
            sink.register(source, dtype)
            self.registrations[sink].add((source, dtype))

        self.sinks.append(sink)
        return sink

    def register(self, system, dtype=None, **kwargs):
        '''
        Register a source system with all the known sinks

        Parameters
        ----------
        system : source.DataSource, source.MultiChanDataSource, or string
            System to register with all the sinks
        dtype : None (deprecated)
            Even if specified, this is overwritten in the 'else:' condition below

        Returns
        -------
        None
        '''
        if hasattr(system, 'name'):
            name = system.name
        elif isinstance(system, str):
            name = system
        else:
            # assume that the system is a class
            print("Inferring name", system.__module__.split("."), type(system))
            name = system.__module__.split(".")[-1]

        if hasattr(system, 'send_to_sinks_dtype'): # used by source.MultiChanDataSource
            dtype = system.send_to_sinks_dtype
        elif hasattr(system, 'source'):
            dtype = system.source.dtype

        self.sources.append((name, dtype))

        for s in self.sinks:
            if (name, dtype) not in self.registrations[s]:
                self.registrations[s].add((name, dtype))
                s.register(name, dtype, **kwargs)

    def send(self, system, data):
        '''
        Send data from the specified 'system' to all sinks which have been registered

        Parameters
        ----------
        system: string
            Name of the system sending the data
        data: np.array
            Generic data to be handled by each sink. Can be a record array, e.g., for task data.

        Returns
        -------
        None
        '''
        for s in self.sinks:
            s.send(system, data)

    def stop(self):
        '''
        Run the 'stop' method of all the registered sinks
        '''
        for s in self.sinks:
            s.stop()

    def __iter__(self):
        '''
        Returns a python iterator to allow looping over all the
        registered sinks, e.g., to send them all the same data
        '''
        for s in self.sinks:
            yield s


class PrintSink(object):
    '''A null sink which directly prints the received data'''
    def __init__(self):
        print("Starting print sink")

    def register(self, name, dtype):
        print(("Registered name %s with dtype %r"%(name, dtype)))

    def send(self, system, data):
        print(("Received %s data: \n%r"%(system, data)))

    def sendMsg(self, msg):
        print(("### MESSAGE: %s"%msg))

    def close(self):
        print("Ended print sink")



'''
Methods for remotely interacting with the sqlite3 database using remote procedure call (RPC)
For example, linking HDF file to a particular task entry.
'''

import os
import time
import json
import shutil
import datetime
from xmlrpc.server import SimpleXMLRPCDispatcher

from django.http import HttpResponse
from django.views.decorators.csrf import csrf_exempt

from .models import TaskEntry, Subject, Calibration, System, DataFile, Decoder

def save_log(idx, log, dbname='default'):
    entry = TaskEntry.objects.using(dbname).get(pk=idx)
    entry.report = json.dumps(log)
    entry.save()

def save_calibration(subject, system, name, params, dbname='default'):
    print(subject, system)
    subj = Subject.objects.using(dbname).get(name=subject)
    sys = System.objects.using(dbname).get(name=system)
    Calibration(subject=subj, system=sys, name=name, params=params).save(using=dbname)

def save_data(curfile, system, entry, move=True, local=True, custom_suffix=None, dbname='default'):
    suffix = dict(emg="hdf", eyetracker="edf", hdf="hdf", plexon="plx", bmi="pkl", bmi_params="npz", juice_log="png", video="avi", optitrack="tak")
    if system in suffix:
        suff = suffix[system]
    else:  # blackrock system saves multiple files (.nev, .ns1, .ns2, etc.)
        if custom_suffix is not None:
            suff = custom_suffix
        else:
            raise Exception('Must provide a custom suffix for system: ' + system)

    sys = System.objects.using(dbname).get(name=system)
    entry = TaskEntry.objects.using(dbname).get(pk=entry)

    now = entry.date
    today = datetime.date(now.year, now.month, now.day)
    tomorrow = today + datetime.timedelta(days=1)

    entries = TaskEntry.objects.using(dbname).filter(date__gte=today, date__lte=tomorrow)
    enums = dict([(e, i) for i, e in enumerate(entries.order_by("date"))])
    num = enums[entry]

    if move:

        # Set the new filename and filepath
        permfile = "{subj}{time}_{num:02}_te{id}.{suff}".format(
            subj=entry.subject.name[:4].lower(),
            time=time.strftime('%Y%m%d'), num=num+1,
            id=entry.id, suff=suff
        )
        if system == 'blackrock2':
            fullname = os.path.join('/storage/rawdata/blackrock', permfile)
            sys_path = '/storage/rawdata/blackrock'
            print('BLACKROCK SYSTEM: ')
        else:
            fullname = os.path.join(sys.path, permfile)
            sys_path = sys.path

        # Move or copy or rsync the file
        if os.path.abspath(sys_path) == os.path.abspath(os.path.split(curfile)[0]):
            print("moving file...")
            os.rename(curfile, fullname) # from sys_path to sys_path
        elif os.path.exists(sys_path) and not os.path.exists(fullname):
            print("copying file...")
            shutil.copy2(curfile, os.path.join(sys_path, permfile)) # from somewhere to sys_path
        else:
            raise ValueError('Will not overwrite existing files')
    else:
        permfile = curfile

    DataFile(local=local, path=permfile, system=sys, entry=entry).save(using=dbname)
    print("Saved datafile for file=%s -> %s, system=%s, id=%d)..." % (curfile, permfile, system, entry.id))

def save_bmi(name, entry, filename, dbname='default'):
    '''
    Save BMI objects to database
    '''
    entry = TaskEntry.objects.using(dbname).get(pk=entry)
    now = entry.date
    today = datetime.date(now.year, now.month, now.day)
    tomorrow = today + datetime.timedelta(days=1)

    entries = TaskEntry.objects.using(dbname).filter(date__gte=today, date__lte=tomorrow)
    enums = dict([(e, i) for i, e in enumerate(entries.order_by("date"))])
    num = enums[entry]

    pklname = "{subj}{time}_{num:02}_{name}.pkl".format(
        subj=entry.subject.name[:4].lower(),
        time=entry.date.strftime('%Y%m%d'),
        num=num, name=name)
    base = System.objects.using(dbname).get(name='bmi').path

    #Make sure decoder name doesn't exist already:
    #Make sure new decoder name doesn't already exist:
    import os.path
    dec_ix = 0

    while os.path.isfile(os.path.join(base, pklname)):
        pklname = "{subj}{time}_{num:02}_{name}_{ix}.pkl".format(
        subj=entry.subject.name[:4].lower(),
        time=entry.date.strftime('%Y%m%d'),
        num=num, name=name,ix=dec_ix)
        dec_ix += 1

    # Some problems with this on windows with permissions
    shutil.copy2(filename, os.path.join(base, pklname))

    Decoder(name=name,entry=entry,path=pklname).save(using=dbname)
    # try:
    #     decoder_entry = Decoder.objects.using(dbname).get(entry=entry)
    # except:
    #     print('too many decoders to list: ')
    #     import dbfunctions as dbfn
    #     d = dbfn.TaskEntry(entry.pk)
    #     d_list = d.get_decoders_trained_in_block()
    #     for d in d_list:
    #         print(f"{d.pk}, {d.name}")
    print("Saved decoder to %s"%os.path.join(base, pklname))

def cleanup(entry, dbname='default'):
    '''
    Final cleanup after a task is finished
    '''
    te = TaskEntry.objects.using(dbname).get(id=entry)
    tries = 3
    while tries > 0:
        if te.make_hdf_self_contained():
            break
        time.sleep(1) # wait for the hdf file to be created
        tries -= 1

def hide_task_entry(entry, dbname='default'):
    te = TaskEntry.objects.using(dbname).get(id=entry)
    te.visible = False
    te.save()



#############################################################################
##### Register functions for remote procedure call from other processes #####
#############################################################################
dispatcher = SimpleXMLRPCDispatcher(allow_none=True)
dispatcher.register_function(save_log, 'save_log')
dispatcher.register_function(save_calibration, 'save_cal')
dispatcher.register_function(save_data, 'save_data')
dispatcher.register_function(save_bmi, 'save_bmi')
dispatcher.register_function(cleanup, 'cleanup')
dispatcher.register_function(hide_task_entry, 'hide_task_entry')

@csrf_exempt
def rpc_handler(request):
    response = HttpResponse(content_type="application/xml")
    response.write(dispatcher._marshaled_dispatch(request.body))
    return response



import multiprocessing as mp
import time
import numpy as np
import queue

class MPCompute(mp.Process):
    """
    Generic class for running computations that occur infrequently
    but take longer than a single BMI loop iteration
    """
    def __init__(self, work_queue, result_queue, fn):
        '''
        Constructor for MPCompute

        Parameters
        ----------
        work_queue : mp.Queue
            Jobs start when an entry is found in work_queue
        result_queue : mp.Queue
            Results of job are placed back onto result_queue

        Returns
        -------
        MPCompute instance
        '''
        # run base constructor
        super(MPCompute, self).__init__()

        self.work_queue = work_queue
        self.result_queue = result_queue
        self.done = mp.Event()
        self.fn = fn

    def _check_for_job(self):
        '''
        Non-blocking check to see if data is present in the input queue
        '''
        try:
            job = self.work_queue.get_nowait()
        except:
            job = None
        return job
        
    def run(self):
        '''
        The main loop. Starts automatically when the process is spawned. See mp.Process.run for additional docs.
        Every 0.5 seconds, check for new computation to carry out

        Parameters
        ----------
        None

        Returns
        -------
        None
        '''
        while not self.done.is_set():
            job = self._check_for_job()

            # unpack the data
            if not (job is None):
                new_params = self.calc(*job[0], **job[1])
                self.result_queue.put(new_params)

            # Pause to lower the process's effective priority
            time.sleep(0.5)

    def calc(self, *args, **kwargs):
        '''
        Run the actual calculation function
        '''
        return self.fn(*args, **kwargs)

    def stop(self):
        '''
        Set the flag to stop the 'while' loop in the 'run' method gracefully
        '''
        self.done.set()


class FuncProxy(object):
    '''
    Wrapper for MPCompute computations running in another process
    '''
    def __init__(self, fn, multiproc=False, waiting_resp=None, init_resp=None, verbose=False):
        self.verbose = verbose
        self.multiproc = multiproc
        if self.multiproc:
            # create the queues
            self.work_queue = mp.Queue()
            self.result_queue = mp.Queue()

            # Instantiate the process
            self.calculator = MPCompute(self.work_queue, self.result_queue, fn)

            # spawn the process
            self.calculator.start()
        else:
            self.fn = fn

        assert waiting_resp in [None, 'prev'], "Unrecognized waiting_resp"
        self.waiting_resp = waiting_resp

        self.prev_result = (init_resp, 0)
        self.prev_input = None
        self.waiting = False

    def reset(self):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        self.prev_input = None

    def _stuff(self):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        try:
            output_data = self.result_queue.get_nowait()
            self.prev_result = output_data
            self.waiting = False
            return output_data, True
        except queue.Empty:
            if self.waiting_resp == None:
                return None
            elif self.waiting_resp == 'prev':
                return self.prev_result, False
        except:
            import traceback
            traceback.print_exc()

    def input_same(self, stuff):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        args, kwargs = stuff
        if self.prev_input == None:
            return False

        args_same = True
        for a1, a2 in zip(args, self.prev_input[0]):
            try: 
                args_same = args_same and np.all(a1 == a2)
            except ValueError:
                args_same = args_same and np.array_equal(a1, a2)

        kwargs_same = list(kwargs.keys()) == list(self.prev_input[1].keys())

        for key1, key2 in zip(list(kwargs.keys()), list(self.prev_input[1].keys())):
            k1 = kwargs[key1]
            k2 = kwargs[key2]
            if key1 == 'q_start': 
                continue
            try:
                kwargs_same = kwargs_same and np.all(k1 == k2)
            except:
                kwargs_same = kwargs_same and np.array_equal(k1, k2)

        return args_same and kwargs_same


    def __call__(self, *args, **kwargs):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        input_data = (args, kwargs)
        input_same_as_last = self.input_same(input_data) #input_data == self.prev_input        
        if self.multiproc:
            if input_same_as_last and not self.waiting:
                return self.prev_result, False

            elif input_same_as_last and self.waiting:
                # Return the new result if it's available, otherwise the previous result
                return self._stuff()

            elif not input_same_as_last:
                if self.verbose: print("queuing job")
                self.work_queue.put(input_data)    
                self.prev_input = input_data
                self.waiting = True
                return self._stuff()
        else:
            if input_same_as_last:
                return self.prev_result, False
            else:
                self.prev_input = input_data
                self.prev_result = self.fn(*args, **kwargs)
                return self.prev_result, True

    def __del__(self):
        '''
        Stop the child process if one was spawned
        '''
        if self.multiproc:
            self.calculator.stop()

class MultiprocShellCommand(mp.Process):
    '''
    Execute a blocking shell command in a separate process
    '''
    def __init__(self, cmd, *args, **kwargs):
        self.cmd = cmd
        self.done = mp.Event()
        super(MultiprocShellCommand, self).__init__(*args, **kwargs)

    def run(self):
        '''
        Docstring
        '''
        import os
        os.popen(self.cmd)
        self.done.set()

    def is_done(self):
        return self.done.is_set()


'''
Features for the eyetracker system
'''

import tempfile
import numpy as np
from riglib import calibrations
from riglib.experiment import traits
from riglib.gpio import ArduinoGPIO
from riglib.oculomatic import oculomatic
from riglib.pupillabs import utils
from built_in_tasks.target_graphics import *
from built_in_tasks.target_capture_task import ScreenTargetCapture
from riglib.stereo_opengl.primitives import AprilTag
from .peripheral_device_features import *

import aopy
import glob
import os

###### CONSTANTS
sec_per_min = 60

class EyeCalibration(traits.HasTraits):
    '''
    Calculates 'calibrated_eye_pos' from 'eye_pos' (added by EyeStreaming feature) by regressing previous
    recorded cursor position and eye position when cursor enters targets. Must specify taskid for calibration. 
    '''

    taskid_for_eye_calibration = traits.Int(0, desc="directory where hdf file lives")
    show_eye_pos = traits.Bool(False, desc="Whether to show eye positions")
    eye_target_calibration = traits.Bool(False, desc="Whether to regress eye positions against target positions")

    def __init__(self, *args, **kwargs): #, start_pos, calibration):
        super(EyeCalibration,self).__init__(*args, **kwargs)
        
        # proc_exp # preprocess cursor data only
        taskid = self.taskid_for_eye_calibration
        hdf_dir = '/storage/hdf'
        hdf_file = glob.glob(os.path.join(hdf_dir, f'*{taskid}*'))[0]
        ecube_file = glob.glob(f'/media/NeuroAcq/*{taskid}*')[0]
        files = {}
        files['hdf'] = hdf_file
        files['ecube'] = ecube_file
        print(files)

        if not self.keyboard_control:
            bmi3d_data, bmi3d_metadata = aopy.preproc.proc_exp(hdf_dir, files, 'hoge', 'hoge', overwrite=True, save_res=False)

            # load raw eye data
            # raw_eye_data, raw_eye_metadata = aopy.preproc.parse_oculomatic(hdf_dir, files, debug=False)
            eye_interp = aopy.data.get_interp_kinematics(bmi3d_data,bmi3d_metadata,datatype='eye',samplerate=bmi3d_metadata['cursor_interp_samplerate'])

            # calculate coefficients to calibrate eye data
            events = bmi3d_data['events']

            if not self.eye_target_calibration:
                self.eye_coeff,_,_,_ = aopy.preproc.calc_eye_calibration\
                    (bmi3d_data['cursor_interp'],bmi3d_metadata['cursor_interp_samplerate'],\
                    eye_interp[:,:4], bmi3d_metadata['cursor_interp_samplerate'],events['timestamp'], events['code'],return_datapoints=True)

            else:
                def get_target_locations(data, target_indices):

                    try:
                        trials = data['trials']
                    except:
                        trials = data['bmi3d_trials']
                    locations = np.nan*np.zeros((len(target_indices), 3))
                    for i in range(len(target_indices)):
                        trial_idx = np.where(trials['index'] == target_indices[i])[0]
                        if len(trial_idx) > 0:
                            locations[i,:] = trials['target'][trial_idx[0]][[0,2,1]] # use x,y,z format
                        else:
                            raise ValueError(f"Target index {target_indices[i]} not found")
                    return np.round(locations,4)

                target_pos = get_target_locations(bmi3d_data, [1,2,3,4,5,6,7,8])
                
                self.eye_coeff, _ = aopy.preproc.calc_eye_target_calibration(eye_interp[:,:4], \
                    bmi3d_metadata['cursor_interp_samplerate'], events['timestamp'], events['code'], target_pos)
            
            print("Calibration complete:", self.eye_coeff)

        # Set up eye cursor
        self.eye_cursor = VirtualCircularTarget(target_radius=.5, target_color=(0., 1., 0., 0.75))
        self.target_location = np.array(self.starting_pos).copy()
        self.calibrated_eye_pos = np.zeros((2,))*np.nan
        for model in self.eye_cursor.graphics_models:
            self.add_model(model)
        
    def init(self):
        self.add_dtype('calibrated_eye', 'f8', (2,))
        super().init()

    #### STATE FUNCTIONS ####
    def _start_wait(self):
        super()._start_wait()

        if self.calc_trial_num() == 0:
            if self.show_eye_pos:
                self.eye_cursor.show()
            else:
                self.eye_cursor.hide()

    def _cycle(self):
        self._update_eye_pos()

        # Do calibration
        ave_pos = self.eye_pos
        if not self.keyboard_control:
            calibrated_pos = aopy.postproc.get_calibrated_eye_data(self.eye_pos[:4],self.eye_coeff)
            ave_pos = np.array([(calibrated_pos[0] + calibrated_pos[2])/2, (calibrated_pos[1] + calibrated_pos[3])/2])
        
        # Save calibration
        self.calibrated_eye_pos = ave_pos
        self.task_data['calibrated_eye'] = ave_pos

        super(EyeStreaming, self)._cycle()

        # Move the eye cursor
        if np.any(np.isnan(self.calibrated_eye_pos)):
            pass
        else:
            self.eye_cursor.move_to_position([self.calibrated_eye_pos[0],0,self.calibrated_eye_pos[1]])
            if self.show_eye_pos:
                self.eye_cursor.show()


class EyeStreaming(traits.HasTraits):
    '''
    Adds eye_data streamed from oculomatic.
    '''

    keyboard_control = traits.Bool(False, desc="Whether to replace eye control with keyboard control")
    eye_labels = traits.Array(value=['le_x', 'le_y', 're_x', 're_y', 'le_diam', 're_diam'], desc="Description of eye data columns")

    hidden_traits = ['eye_labels']

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Visualize eye positions
        if self.keyboard_control:
            self.eye_data = Eye([0,0])
            self.eye_pos = np.zeros((2,))*np.nan
        else:
            from riglib import source
            from riglib.oculomatic import System
            self.eye_data = source.DataSource(System)
            self.eye_pos = np.zeros((6,))*np.nan


    def init(self):
        if self.keyboard_control:
            self.add_dtype('eye', 'f8', (2,))
        else:
            self.add_dtype('eye', 'f8', (6,))
        super().init()

    def run(self):
        '''
        Code to execute immediately prior to the beginning of the task FSM executing, or after the FSM has finished running. 
        See riglib.experiment.Experiment.run(). This 'run' method starts the motiondata source and stops it after the FSM has finished running
        '''
        if not self.keyboard_control:
            self.eye_data.start()
        try:
            super().run()
        finally:
            if not self.keyboard_control:
                print("Stopping streaming eye data")
                self.eye_data.stop()

    def _update_eye_pos(self):
        if not self.keyboard_control:
            eye_pos = self.eye_data.get() # This is (n,6) array of new values since we last checked
            if eye_pos.ndim < 2 or eye_pos.size == 0:
                eye_pos = np.zeros((6,))*np.nan
            else:
                eye_pos = eye_pos[-1,:] # the most recent position
        else:
            eye_pos = self.eye_data.get() # A list of lists of of x,y keyboard pos
            eye_pos = eye_pos[0]
        self.eye_pos = eye_pos
        self.task_data['eye'] = eye_pos

    def _cycle(self):
        self._update_eye_pos()
        super()._cycle()


class EyeConstrained(ScreenTargetCapture):
    '''
    Add a penalty state when subjects looks away. Only tested in center-out task.
    '''

    fixation_dist = traits.Float(2.5, desc="Distance from center that is considered a broken fixation")
    fixation_penalty_time = traits.Float(0., desc="Time in fixation penalty state")
    fixation_target_color = traits.OptionsList("cyan", *target_colors, desc="Color of the center target under fixation state", bmi3d_input_options=list(target_colors.keys()))
    
    status = dict(
        wait = dict(start_trial="target"),
        target = dict(timeout="timeout_penalty",gaze_target="fixation"),
        fixation = dict(enter_target="hold", fixation_break="target"),
        hold = dict(leave_target="hold_penalty", hold_complete="delay", fixation_break="fixation_penalty"),
        delay = dict(leave_target="delay_penalty", delay_complete="targ_transition", fixation_break="fixation_penalty"),
        targ_transition = dict(trial_complete="reward", trial_abort="wait", trial_incomplete="target"),
        timeout_penalty = dict(timeout_penalty_end="targ_transition", end_state=True),
        hold_penalty = dict(hold_penalty_end="targ_transition", end_state=True),
        delay_penalty = dict(delay_penalty_end="targ_transition", end_state=True),
        fixation_penalty = dict(fixation_penalty_end="targ_transition",end_state=True),
        reward = dict(reward_end="wait", stoppable=False, end_state=True)
    )
 
    def _test_gaze_target(self,ts):
        '''
        Check whether eye positions are within the fixation distance
        Only apply this to the first target (1st target)
        '''
        if self.target_index <= 0:     
            d = np.linalg.norm(self.calibrated_eye_pos)
            return d < self.fixation_dist
        else:
            return True
        
    def _test_fixation_break(self,ts):
        '''
        Triggers the fixation_penalty state when eye positions are outside fixation distance
        Only apply this to the first hold and delay period
        '''
        if self.target_index <= 0:   
            d = np.linalg.norm(self.calibrated_eye_pos)
            return (d > self.fixation_dist) or self.pause
        else:
            return self.pause
    
    def _test_fixation_penalty_end(self,ts):
        # d = np.linalg.norm(self.calibrated_eye_pos)
        return (ts > self.fixation_penalty_time) # (d < self.fixation_dist) and 
    
    def _start_wait(self):
        super()._start_wait()
        self.num_fixation_state = 0 # Initializa fixation state

    def _start_target(self):
        if self.num_fixation_state == 0:
            super()._start_target() # target index shouldn't be incremented after fixation break loop
        else:
            self.sync_event('FIXATION', 0)
            self.targets[0].reset() # reset target color after fixation break

    def _start_fixation(self):
        self.num_fixation_state = 1
        self.targets[0].sphere.color = target_colors[self.fixation_target_color] # change target color in fixation state
        if self.target_index == 0:
            self.sync_event('FIXATION', 1)
    
    def _start_timeout_penalty(self):
        super()._start_timeout_penalty()
        self.num_fixation_state = 0

    def _start_hold(self):
        super()._start_hold()
        self.num_fixation_state = 0 # because target state comes again after hold state in a trial

    def _start_fixation_penalty(self):
        self._increment_tries()
        self.sync_event('FIXATION_PENALTY') 

        # Hide targets
        for target in self.targets:
            target.hide()
            target.reset()

    def _end_fixation_penalty(self):
        self.sync_event('TRIAL_END')

class PupilLabStreaming(traits.HasTraits):
    '''
    Adds eye_data from pupil labs. Optionally displays AprilTag markers on the screen for
    surface tracking. Requires a task with the Window feature enabled.
    '''

    surface_marker_size = traits.Float(2., desc="Size in cm of apriltag surface markers")
    surface_marker_count = traits.Int(0, desc="How many surface markers to draw")
    eye_labels = traits.Array(value=[
        'gaze_x', 'gaze_y', 'gaze_z', 
        'norm_x', 'norm_y', 
        'timestamp', 
        're_x', 're_y', 
        're_diam', 
        'le_x', 'le_y', 
        'le_diam'
        ], 
        desc="Description of eye data columns")

    hidden_traits = ['eye_labels']

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        
        # Add apriltag models
        centers = utils.calculate_square_positions(self.screen_half_height, self.window_size, 
                                                   self.surface_marker_count, self.surface_marker_size)
        for id, (x, z) in enumerate(centers):
            tag = AprilTag(id, self.surface_marker_size).translate(x, 0, z)
            self.add_model(tag)

        # Visualize eye positions
        from riglib import source
        from riglib.pupillabs import System, NoSurfaceTracking
        if self.surface_marker_count > 0:
            self.eye_data = source.DataSource(System)
        else:
            self.eye_data = source.DataSource(NoSurfaceTracking)
        self.eye_pos = np.zeros((8,))*np.nan

    def init(self):
        self.add_dtype('eye', 'f8', (12,))
        super().init()

    def run(self):
        '''
        Code to execute immediately prior to the beginning of the task FSM executing, or after the FSM has finished running. 
        See riglib.experiment.Experiment.run(). This 'run' method starts the motiondata source and stops it after the FSM has finished running
        '''
        self.eye_data.start()
        try:
            super().run()
        finally:
            print("Stopping streaming eye data")
            self.eye_data.stop()

    def _start_None(self):
        '''
        Code to run before the 'None' state starts (i.e., the task stops)
        '''
        self.eye_data.stop()
        super()._start_None()

    def _update_eye_pos(self):
        eye_pos = self.eye_data.get() # This is (n,11) array of new values since we last checked
        if eye_pos.ndim < 2 or eye_pos.size == 0:
            eye_pos = np.zeros((12,))*np.nan
        else:
            eye_pos = eye_pos[-1,:] # the most recent position
        self.eye_pos = eye_pos
        self.task_data['eye'] = eye_pos

    def _cycle(self):
        self._update_eye_pos()
        super()._cycle()

'''
Old code not currently used in aolab
'''
class EyeData(traits.HasTraits):
    '''
    Pulls data from the eyetracking system and make it available on self.eyedata
    '''
    def init(self):
        '''
        Secondary init function. See riglib.experiment.Experiment.init()
        Prior to starting the task, this 'init' sets up the 'eyedata' DataSource and registers it with the 
        SinkRegister so that the data gets saved to file as it is collected.
        '''
        from riglib import source
        from riglib import sink
        sink_manager = sink.SinkManager.get_instance()

        src, ekw = self.eye_source
        #f = open('/home/helene/code/bmi3d/log/eyetracker', 'a')
        self.eyedata = source.DataSource(src, **ekw)
        sink_manager.register(self.eyedata)
        f.write('instantiated source\n')
        super(EyeData, self).init()
        #f.close()
    
    @property
    def eye_source(self):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        from riglib import eyetracker
        return eyetracker.System, dict()

    def run(self):
        '''
        Code to execute immediately prior to the beginning of the task FSM executing, or after the FSM has finished running. 
        See riglib.experiment.Experiment.run(). This 'run' method starts the 'eyedata' source and stops it after the FSM has finished running
        '''
        #f = open('/home/helene/code/bmi3d/log/eyetracker', 'a')
        self.eyedata.start()
        #f.write('started eyedata\n')
        #f.close()
        try:
            super(EyeData, self).run()
        finally:
            self.eyedata.stop()
    
    def join(self):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        self.eyedata.join()
        super(EyeData, self).join()
    
    def _start_None(self):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        self.eyedata.pause()
        self.eyefile = tempfile.mktemp()
        print("retrieving data from eyetracker...")
        self.eyedata.retrieve(self.eyefile)
        print("Done!")
        self.eyedata.stop()
        super(EyeData, self)._start_None()
    
    def set_state(self, state, **kwargs):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        self.eyedata.sendMsg(state)
        super(EyeData, self).set_state(state, **kwargs)

    def cleanup(self, database, saveid, **kwargs):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        
        super(EyeData, self).cleanup(database, saveid, **kwargs)
        dbname = kwargs['dbname'] if 'dbname' in kwargs else 'default'
        if dbname == 'default':
            database.save_data(self.eyefile, "eyetracker", saveid)
        else:
            database.save_data(self.eyefile, "eyetracker", saveid, dbname=dbname)

class SimulatedEyeData(EyeData):
    '''Simulate an eyetracking system using a series of fixations, with saccades interpolated'''
    fixations = traits.Array(value=[(0,0), (-0.6,0.3), (0.6,0.3)], desc="Location of fixation points")
    fixation_len = traits.Float(0.5, desc="Length of a fixation")

    @property
    def eye_source(self):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        from riglib import eyetracker
        return eyetracker.Simulate, dict(fixations= self.fixations)

    def _cycle(self):
        '''
        Docstring
        basically, extract the data and do something with it


        Parameters
        ----------

        Returns
        -------
        '''
        #retrieve data
        data_temp = self.eyedata.get()

        #send the data to sinks
        if data_temp is not None:
            self.sinks.send(self.eyedata.name, data_temp)

        super(SimulatedEyeData, self)._cycle()

class CalibratedEyeData(EyeData):
    '''Filters eyetracking data with a calibration profile'''
    cal_profile = traits.Instance(calibrations.EyeProfile)

    def __init__(self, *args, **kwargs):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        super(CalibratedEyeData, self).__init__(*args, **kwargs)
        self.eyedata.set_filter(self.cal_profile)

class FixationStart(CalibratedEyeData):
    '''Triggers the start_trial event whenever fixation exceeds *fixation_length*'''
    fixation_length = traits.Float(2., desc="Length of fixation required to start the task")
    fixation_dist = traits.Float(50., desc="Distance from center that is considered a broken fixation")

    def __init__(self, *args, **kwargs):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        super(FixationStart, self).__init__(*args, **kwargs)
        self.status['wait']['fixation_break'] = "wait"
        self.log_exclude.add(("wait", "fixation_break"))
    
    def _start_wait(self):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        self.eyedata.get()
        super(FixationStart, self)._start_wait()

    def _test_fixation_break(self, ts):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        return (np.sqrt((self.eyedata.get()**2).sum(1)) > self.fixation_dist).any()
    
    def _test_start_trial(self, ts):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        return ts > self.fixation_length



from .gpio import ArduinoGPIO
from multiprocessing import Process
from riglib import singleton
#from pyfirmata import Arduino, util
import traceback
import time
import os

log_path = os.path.join(os.path.dirname(__file__), '../log/reward.log')

class Basic(singleton.Singleton):

    __instance = None

    def __init__(self):
        super().__init__()
        self.board = ArduinoGPIO() # let the computer find the arduino. this won't work with more than one arduino!
        self.board.write(12,1)

    def dispense(self):  # call this function to drain the reward system
        """
        this function is called from the webserver in ajax.reward_drain
        """
        self.board.write(12, 0)          #low
        time.sleep(0.02)
        self.board.write(12, 1)            #high

def open():
    try:
        reward = Basic.get_instance()
        return reward
    except:
        print("Reward system not found/ not active")
        import traceback
        import os
        import builtins
        traceback.print_exc()

import requests
class RemoteReward():

    def __init__(self):

        self.hostName = "192.168.0.200"
        self.serverPort = 8080

    def trigger(self):
        url = f"http://{self.hostName}:{self.serverPort}"
        try:
            requests.post(url, timeout=3)
        except:
            traceback.print_exc()



'''
Base code for including the 'eyetracker' features in experiments
'''


import time
import itertools
import numpy as np
try:
    import pylink
except ImportError:
    print("Couldn't find eyetracker module")

class Simulate(object):
    '''
    Feature (task add-on) to simulate the eyetracker.
    '''
    update_freq = 500
    dtype = np.dtype((float, (2,)))

    def __init__(self, fixations=[(0,0), (-0.6,0.3), (0.6,0.3)], isi=500, slen=15):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        from scipy.interpolate import interp1d
        flen = list(range(len(fixations)+1))
        t = list(itertools.chain(*[(i*isi + slen*i, (i+1)*isi + slen*i) for i in flen]))[:-1]
        xy = np.append(np.tile(fixations, (1, 2)).reshape(-1, 2), [fixations[0]], axis=0)
        self.mod = t[-1] / 1000.
        self.interp = interp1d(np.array(t)/1000., xy, kind='linear', axis=0)
        self.fixations = fixations
        self.isi = isi

    def start(self):

        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        print("eyetracker.simulate.start()")
        self.stime = time.time()
    
    def retrieve(self, filename):
        '''
        for sim, there is no need to retrieve an file

        Parameters
        ----------

        Returns
        -------
        '''
        pass

    def get(self):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        time.sleep(1./self.update_freq)

        data = self.interp((time.time() - self.stime) % self.mod) + np.random.randn(2)*.01
        #expand dims
        data_2 = np.expand_dims(data, axis = 0)
        return data_2

    def stop(self):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        return

    def sendMsg(self, msg):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        pass

class System(object):
    '''
    System representing the EyeLink eyetracker. Compatible with riglib.source.DataSource
    '''
    update_freq = 500
    dtype = np.dtype((float, (2,)))

    def __init__(self, address='10.0.0.2'):
        '''
        Constructor for the System representing the EyeLink eyetracker

        Parameters
        ----------
        address: IP address string 
            IP address of the EyeLink host machine

        Returns
        -------
        System instance
        '''
        self.tracker = pylink.EyeLink(address)
        self.tracker.setOfflineMode()
    
    def start(self, filename=None):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        print("eyetracker.System.start()")
        self.filename = filename
        if filename is None:
            self.filename = "%s.edf"%time.strftime("%Y%m%d") #%Y-%m-%d_%I:%M:%p
        self.tracker.openDataFile(self.filename)
        # pylink.beginRealTimeMode(100)
        print("\n\ntracker.startRecording")
        self.tracker.startRecording(1,0,1,0)

    def stop(self):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        self.tracker.stopRecording()
        pylink.endRealTimeMode()
    
    def get(self):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        samp = self.tracker.getNextData()
        while samp != pylink.SAMPLE_TYPE:
            time.sleep(.001)
            samp = self.tracker.getNextData()
        try:
            data = np.array(self.tracker.getFloatData().getLeftEye().getGaze())
            if data.sum() < -1e4:
                return np.array([np.nan, np.nan])
        except:
            return np.array([np.nan, np.nan])
            
        return data
        
    def set_filter(self, filt):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        self.filter = filt
    
    def retrieve(self, filename):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        self.tracker.setOfflineMode()
        pylink.msecDelay(1)
        self.tracker.closeDataFile()
        self.tracker.receiveDataFile(self.filename, filename)
    
    def sendMsg(self, msg):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        self.tracker.sendMessage(msg)

    def __del__(self):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        self.tracker.close()



'''
Closed-loop decoder adaptation (CLDA) classes. There are two types of classes,
"Learners" and "Updaters". Learners implement various methods to estimate the
"intended" BMI movements of the user. Updaters implement various method for 
updating 
'''
import multiprocessing as mp
import numpy as np
from . import kfdecoder, ppfdecoder, train
import re
import copy

from utils.angle_utils import *
from riglib.mp_calc import MPCompute

inv = np.linalg.inv

try:
    from numpy.linalg import lapack_lite
    lapack_routine = lapack_lite.dgesv
except:
    pass
    
def fast_inv(A):
    '''
    This method represents a way to speed up matrix inverse computations when 
    several independent matrix inverses of all the same shape must be taken all
    at once. This is used by the PPFContinuousBayesianUpdater. Without this method,
    the updates could not be performed in real time with ~30 cells (compute complexity 
    is linear in the number of units, so it is possible that fewer units would not
    have had this issue).

    Code stolen from: 
    # http://stackoverflow.com/questions/11972102/is-there-a-way-to-efficiently-invert-an-array-of-matrices-with-numpy 
    '''
    b = np.identity(A.shape[2], dtype=A.dtype)

    n_eq = A.shape[1]
    n_rhs = A.shape[2]
    pivots = np.zeros(n_eq, np.intc)
    identity  = np.eye(n_eq)
    def lapack_inverse(a):
        b = np.copy(identity)
        pivots = np.zeros(n_eq, np.intc)
        results = lapack_lite.dgesv(n_eq, n_rhs, a, n_eq, pivots, b, n_eq, 0)
        if results['info'] > 0:
            raise np.LinAlgError('Singular matrix')
        return b

    return np.array([lapack_inverse(a) for a in A])

def slow_inv(A):
    return np.array([np.linalg.inv(a) for a in A])

##############################################################################
## Learners
##############################################################################
class Learner(object):
    '''
    Classes for estimating the 'intention' of the BMI operator, inferring the intention from task goals.
    '''
    def __init__(self, batch_size, *args, **kwargs):
        '''
        Instantiate a Learner for estimating intention during CLDA

        Parameters
        ----------
        batch_size: int
            number of samples used to estimate each new decoder parameter setting
        done_states: list of strings, optional
            states of the task which end a batch, regardless of the length of the batch. default = []
        reset_states: list of strings, optional
            states of the task which, if encountered, reset the batch regardless of its length. default = []

        '''
        self.done_states = kwargs.pop('done_states', [])
        self.reset_states = kwargs.pop('reset_states', [])
        print("Reset states for learner: ")
        print(self.reset_states)
        print("Done states for learner: ")
        print(self.done_states)        
        self.batch_size = batch_size
        self.passed_done_state = False
        self.enabled = True
        self.input_state_index = 0
        self.reset()

    def disable(self):
        '''Set a flag to disable forming intention estimates from new incoming data'''
        self.enabled = False

    def enable(self):
        '''Set a flag to enable forming intention estimates from new incoming data'''
        self.enabled = True

    def reset(self):
        '''Reset the lists of saved intention estimates and corresponding neural data'''
        self.kindata = []
        self.neuraldata = []
        self.obs_value = []

    def __call__(self, spike_counts, decoder_state, target_state, decoder_output, task_state, state_order=None, **kwargs):
        """
        Calculate the intended kinematics and pair with the neural data

        Parameters
        ----------
        spike_counts : np.mat of shape (K, 1)
            Neural observations used to decode 'decoder_state'
        decoder_state : np.mat of shape (N, 1)
            State estimate output from the decoder
        target_state : np.mat of shape (N, 1)
            For the current time, this is the optimal state for the Decoder as specified by the task
        decoder_output : np.mat of shape (N, 1)
            ... this seems like the same as decoder_state
        task_state : string
            Name of the task state; some learners (e.g., the cursorGoal learner) have different intention estimates depending on the phase of the task/trial
        state_order : np.ndarray of shape (N,), optional
            Order of each state in the decoder; see riglib.bmi.state_space_models.State
        **kwargs: dict
            Optional keyword arguments for the 'value' calculator

        Returns
        -------
        None
        """
        if task_state in self.reset_states:
            print("resetting CLDA batch")
            self.reset()

        int_kin = self.calc_int_kin(decoder_state, target_state, decoder_output, task_state, state_order=state_order)
        obs_value = self.calc_value(decoder_state, target_state, decoder_output, task_state, state_order=state_order, **kwargs)
        
        if self.passed_done_state and self.enabled:
            if task_state in ['hold', 'target']:
                self.passed_done_state = False

        if self.enabled and not self.passed_done_state and int_kin is not None:
            self.kindata.append(int_kin)
            self.neuraldata.append(spike_counts)
            self.obs_value.append(obs_value)

            if task_state in self.done_states:
                self.passed_done_state = True

    def calc_value(self, *args, **kwargs):
        '''
        Calculate a "value", i.e. a usefulness, for a particular observation. 
        Can override in child classe for RL-style updates, but a priori all observations are equally informative
        '''
        return 1.

    def is_ready(self):
        '''
        Returns True if the collected estimates of the subject's intention are ready for processing into new decoder parameters
        '''
        _is_ready = len(self.kindata) >= self.batch_size or ((len(self.kindata) > 0) and self.passed_done_state)
        return _is_ready

    def get_batch(self):
        '''
        Returns all the data from the last 'batch' of obserations of intended kinematics and neural decoder inputs
        '''
        kindata = np.hstack(self.kindata)
        neuraldata = np.hstack(self.neuraldata)
        self.reset()
        return dict(intended_kin=kindata, spike_counts=neuraldata)
        # return kindata, neuraldata

class DumbLearner(Learner):
    '''
    A learner that never learns anything. Used to make non-adaptive BMI tasks interface the same as CLDA tasks.
    '''
    def __init__(self, *args, **kwargs):
        '''
        Constructor for DumbLearner

        Parameters
        ----------
        args, kwargs: positional and keyword arguments
            Ignored, none are needed

        Returns
        -------
        DumbLearner instance
        '''
        self.enabled = False
        self.input_state_index = 0

    def __call__(self, *args, **kwargs):
        """
        Do nothing; hence the name of the class

        Parameters
        ----------
        args, kwargs: positional and keyword arguments
            Ignored, none are needed

        Returns
        -------
        None
        """
        pass

    def is_ready(self):
        '''DumbLearner is never ready to tell you what it learnerd'''
        return False

    def get_batch(self):
        '''DumbLearner never has any 'batch' data to retrieve'''
        raise NotImplementedError

class FeedbackControllerLearner(Learner):
    '''
    An intention estimator where the subject is assumed to operate like a state feedback controller
    '''
    def __init__(self, batch_size, fb_ctrl, *args, **kwargs):
        self.fb_ctrl = fb_ctrl
        self.style = kwargs.pop('style', 'mixing')
        super(FeedbackControllerLearner, self).__init__(batch_size, *args, **kwargs)

    def calc_int_kin(self, current_state, target_state, decoder_output, task_state, state_order=None):
        """
        Used by __call__ to figure out the next state vector to pair to the neural activity in the batch.

        Parameters
        ----------
        [same OFCLearner.calc_int_kin]
        current_state : np.mat of shape (N, 1)
            State estimate output from the decoder.
        target_state : np.mat of shape (N, 1)
            For the current time, this is the optimal state for the Decoder as specified by the task
        decoder_output : np.mat of shape (N, 1)
            State estimate output from the decoder, after the current observations (may be one step removed from 'current_state')
        task_state : string
            Name of the task state; some learners (e.g., the cursorGoal learner) have different intention estimates depending on the phase of the task/trial
        state_order : np.ndarray of shape (N,), optional
            Order of each state in the decoder; see riglib.bmi.state_space_models.State

        Returns
        -------
        np.mat of shape (N, 1)
            Optimal next state to pair to neural activity
        """        
        try:
            if self.style == 'additive':
                output = self.fb_ctrl(current_state, target_state, mode=task_state)
            elif self.style == 'mixing':
                output = self.fb_ctrl.calc_next_state(current_state, target_state, mode=task_state)
            return output
        except:
            # Key errors happen when the feedback controller doesn't have a policy for the current task state
            import traceback
            traceback.print_exc()
            return None

class OFCLearner(Learner):
    '''
    An intention estimator where the subject is assumed to operate like a muiti-modal LQR controller
    '''
    def __init__(self, batch_size, A, B, F_dict, *args, **kwargs):
        '''
        Constructor for OFCLearner

        Parameters
        ----------
        batch_size : int
            size of batch of samples to pass to the Updater to estimate new decoder parameters
        A : np.mat 
            State transition matrix of the modeled discrete-time system
        B : np.mat 
            Control input matrix of the modeled discrete-time system
        F_dict : dict
            Keys match names of task states, values are feedback matrices (size n_inputs x n_states)
        *args : additional comma-separated args
            Passed to super constructor
        **kwargs : additional keyword args
            Passed to super constructor

        Returns
        -------
        OFCLearner instance
        '''
        super(OFCLearner, self).__init__(batch_size, *args, **kwargs)
        self.B = B
        self.F_dict = F_dict
        self.A = A

    def calc_int_kin(self, current_state, target_state, decoder_output, task_state, state_order=None):
        '''
        Calculate intended kinematics as 
            x_t^{int} = A*x_t + B*F(x^* - x_t)

        Parameters
        ----------
        [same FeedbackControllerLearner.calc_int_kin]
        current_state : np.mat of shape (N, 1)
            State estimate output from the decoder.
        target_state : np.mat of shape (N, 1)
            For the current time, this is the optimal state for the Decoder as specified by the task
        decoder_output : np.mat of shape (N, 1)
            State estimate output from the decoder, after the current observations (may be one step removed from 'current_state')
        task_state : string
            Name of the task state; some learners (e.g., the cursorGoal learner) have different intention estimates depending on the phase of the task/trial
        state_order : np.ndarray of shape (N,), optional
            Order of each state in the decoder; see riglib.bmi.state_space_models.State

        Returns
        -------
        np.mat of shape (N, 1)
            Estimate of intended next state for BMI
        '''
        try:
            current_state = np.mat(current_state).reshape(-1,1)
            target_state = np.mat(target_state).reshape(-1,1)
            F = self.F_dict[task_state]
            A = self.A
            B = self.B

            return A*current_state + B*F*(target_state - current_state)
        except KeyError:
            return None

class OFCLearnerRotateIntendedVelocity(OFCLearner):
    # Only intended for 3D cursor control with this state vector: [Px, Py, Pz, Vx, Vy, Vz, 1]
    def calc_int_kin(self, current_state, target_state, decoder_output, task_state, state_order=None):
        '''
        Calculate intended kinematics as 
            x_t^{int} = A*x_t + B*F(x^* - x_t)

        Parameters
        ----------
        [same FeedbackControllerLearner.calc_int_kin]
        current_state : np.mat of shape (N, 1)
            State estimate output from the decoder.
        target_state : np.mat of shape (N, 1)
            For the current time, this is the optimal state for the Decoder as specified by the task
        decoder_output : np.mat of shape (N, 1)
            State estimate output from the decoder, after the current observations (may be one step removed from 'current_state')
        task_state : string
            Name of the task state; some learners (e.g., the cursorGoal learner) have different intention estimates depending on the phase of the task/trial
        state_order : np.ndarray of shape (N,), optional
            Order of each state in the decoder; see riglib.bmi.state_space_models.State

        Returns
        -------
        np.mat of shape (N, 1)
            Estimate of intended next state for BMI
        '''
        try:        
            current_position = current_state[:3] # assumes the first half of states are position (besides offset)
            target_position = target_state[:3]
            target_velocity_norm = (target_position - current_position)/np.linalg.norm((target_position - current_position)) # Unit velocity vector towards target

            target_state = np.ones(7)
            target_state[:3] = current_position # Set target state position to be the same as the current state position
            target_state[3:6] = target_velocity_norm * np.linalg.norm(current_state[3:6])


            current_state = np.mat(current_state).reshape(-1,1)
            target_state = np.mat(target_state).reshape(-1,1)
            F = self.F_dict[task_state]
            A = self.A
            B = self.B

            return A*current_state + B*F*(target_state - current_state)
        except KeyError:
            return None


class RegexKeyDict(dict):
    '''
    Dictionary where key matching applies regular expressions in addition to exact matches
    '''
    def __getitem__(self, key):
        '''
        Lookup key in dictionary by finding exactly one dict key which, by regex, matches the input argument 'key'
        '''
        keys = list(self.keys())
        matching_keys = [x for x in keys if re.match(x, key)]
        if len(matching_keys) == 0:
            raise KeyError("No matching keys were found!")
        elif len(matching_keys) > 1:
            raise ValueError("Multiple keys match!")
        else:
            return super(RegexKeyDict, self).__getitem__(matching_keys[0])

    def __contains__(self, key):
        '''
        Determine if a key is in the dictionary using regular expression matching
        '''
        keys = list(self.keys())
        matching_keys = [x for x in keys if re.match(x, key)]
        if len(matching_keys) == 0:
            return False
        elif len(matching_keys) > 1:
            raise ValueError("Multiple keys match!")
        else:
            return True        

##############################################################################
## Updaters
##############################################################################
class Updater(object):
    '''
    Wrapper for MPCompute computations running in another process
    '''
    def __init__(self, fn, multiproc=False, verbose=False):
        self.verbose = verbose
        self.multiproc = multiproc
        if self.multiproc:
            # create the queues
            self.work_queue = mp.Queue()
            self.result_queue = mp.Queue()

            # Instantiate the process
            self.calculator = MPCompute(self.work_queue, self.result_queue, fn)

            # spawn the process
            self.calculator.start()
        else:
            self.fn = fn

        self._result = None
        self.waiting = False

    def init(self, decoder):
        pass

    def __call__(self, *args, **kwargs):
        input_data = (args, kwargs)
        if self.multiproc:
            if self.verbose: print("queuing job")
            self.work_queue.put(input_data)    
            self.prev_input = input_data
            self.waiting = True
        else:
            self._result = self.fn(*args, **kwargs)

    def get_result(self):
        if self.multiproc:
            try:
                output_data = self.result_queue.get_nowait()
                self.prev_result = output_data
                self.waiting = False
                return output_data
            except Queue.Empty:
                return None
            except:
                import traceback
                traceback.print_exc()
        else:
            if self._result is not None:
                res = self._result
            else:
                res = None
            self._result = None
            return res

    def __del__(self):
        '''
        Stop the child process if one was spawned
        '''
        if self.multiproc:
            self.calculator.stop()

class PPFContinuousBayesianUpdater(Updater):
    '''
    Adapt the parameters of a PPFDecoder using an HMM to implement a gradient-descent type parameter update.

    (currently only works for PPFs which do not also include the self-history or correlational elements)

    See Shanechi and Carmena, "Optimal feedback-controlled point process decoder for 
    adaptation and assisted training in brain-machine interfaces", IEEE EMBC, 2014
    for mathematical details
    '''
    update_kwargs = dict()
    def __init__(self, decoder, units='cm', param_noise_scale=1., param_noise_variances=None):
        '''
        Constructor for PPFContinuousBayesianUpdater

        Parameters
        ----------
        decoder : bmi.ppfdecoder.PPFDecoder instance
            Should have a 'filt' attribute which is a PointProcessFilter instance
        units : string
            Docstring
        param_noise_scale : float
            Multiplicative factor to increase the parameter "process noise". Higher values result in faster but less stable parameter convergence.
        '''
        super(PPFContinuousBayesianUpdater, self).__init__(self.calc, multiproc=False)

        self.n_units = decoder.filt.C.shape[0]
        if param_noise_variances == None:
            if units == 'm':
                vel_gain = 1e-4
            elif units == 'cm':
                vel_gain = 1e-8

            print("Updater param noise scale %g" % param_noise_scale)
            vel_gain *= param_noise_scale
            param_noise_variances = np.array([vel_gain*0.13, vel_gain*0.13, 1e-4*0.06/50])
        self.W = np.tile(np.diag(param_noise_variances), [self.n_units, 1, 1])


        self.P_params_est = self.W.copy()

        self.neuron_driving_state_inds = np.nonzero(decoder.drives_neurons)[0]
        self.neuron_driving_states = list(np.take(decoder.states, np.nonzero(decoder.drives_neurons)[0]))
        self.n_states = len(decoder.states)
        self.full_size = len(decoder.states)

        self.dt = decoder.filt.dt
        self.beta_est = np.array(decoder.filt.C)

    def calc(self, intended_kin=None, spike_counts=None, decoder=None, **kwargs):
        '''    Docstring    '''

        if (intended_kin is None) or (spike_counts is None) or (decoder is None):
            raise ValueError("must specify intended_kin, spike_counts and decoder objects for the updater to work!")        

        if 0:
            print(np.array(intended_kin).ravel())

        int_kin_full = intended_kin
        spike_obs_full = spike_counts
        n_samples = int_kin_full.shape[1]

        # Squash any observed spike counts which are greater than 1
        spike_obs_full[spike_obs_full > 1] = 1
        for k in range(n_samples):
            spike_obs = spike_obs_full[:,k]
            int_kin = int_kin_full[:,k]

            beta_est = self.beta_est[:,self.neuron_driving_state_inds]
            int_kin = np.asarray(int_kin).ravel()[self.neuron_driving_state_inds]
            Loglambda_predict = np.dot(int_kin, beta_est.T)
            rates = np.exp(Loglambda_predict)
            if np.any(rates > 1):
                print('rates > 1!')
                rates[rates > 1] = 1
            unpred_spikes = np.asarray(spike_obs).ravel() - rates

            C_xpose_C = np.outer(int_kin, int_kin)

            self.P_params_est += self.W
            try:
                P_params_est_inv = fast_inv(self.P_params_est)
            except:
                P_params_est_inv = slow_inv(self.P_params_est)
            L = np.dstack([rates[c] * C_xpose_C for c in range(self.n_units)]).transpose([2,0,1])

            try:
                self.P_params_est = fast_inv(P_params_est_inv + L)
            except:
                self.P_params_est = slow_inv(P_params_est_inv + L)

            beta_est += (unpred_spikes * np.dot(int_kin, self.P_params_est).T).T

            # store beta_est
            self.beta_est[:, self.neuron_driving_state_inds] = beta_est

        return {'filt.C': np.mat(self.beta_est.copy())}

class KFRML(Updater):
    '''
    Calculate updates for KF parameters using the recursive maximum likelihood (RML) method
    See (Dangi et al, Neural Computation, 2014) for mathematical details.
    '''
    update_kwargs = dict(steady_state=False)
    def __init__(self, batch_time, half_life, adapt_C_xpose_Q_inv_C=True, regularizer=None):
        '''
        Constructor for KFRML

        Parameters
        ----------
        batch_time : float
            Size of data batch to use for each update. Specify in seconds.
        half_life : float 
            Amount of time (in seconds) before parameters are half-overwritten by new data.
        adapt_C_xpose_Q_inv_C : bool
            Flag specifying whether to update the decoder property C^T Q^{-1} C, which 
            defines the feedback dynamics of the final closed-loop system if A and W are known
        regularizer: float
            Defines lambda regularizer to use in calculation of C matrix : C = (X*X.T + lambda*eye).I * (X*Y)

        Returns
        -------
        KFRML instance
        '''
        super(KFRML, self).__init__(self.calc, multiproc=False)
        self.batch_time = batch_time
        self.half_life = half_life
        self.rho = np.exp(np.log(0.5) / (self.half_life/batch_time))
        self.adapt_C_xpose_Q_inv_C = adapt_C_xpose_Q_inv_C
        self.regularizer = regularizer
        self._new_params = None

    @staticmethod
    def compute_suff_stats(hidden_state, obs, include_offset=True):
        '''
        Calculate initial estimates of the parameter sufficient statistics used in the RML update rules

        Parameters
        ----------
        hidden_state : np.ndarray of shape (n_states, n_samples)
            Examples of the hidden state x_t taken from training seed data.  
        obs : np.ndarray of shape (n_features, n_samples)
            Multiple neural observations paired with each of the hidden state examples
        include_offset : bool, optional
            If true, a state of all 1's is added to the hidden_state to represent mean offsets. True by default

        Returns
        -------
        R : np.ndarray of shape (n_states, n_states)
            Proportional to covariance of the hidden state samples 
        S : np.ndarray of shape (n_features, n_states)
            Proportional to cross-covariance between 
        T : np.ndarray of shape (n_features, n_features)
            Proportional to covariance of the neural observations
        ESS : float
            Effective number of samples. In the initialization, this is just the 
            dimension of the array passed in, but the parameter can become non-integer 
            during the update procedure as old parameters are "forgotten".
        '''
        assert hidden_state.shape[1] == obs.shape[1]
    
        if isinstance(hidden_state, np.ma.core.MaskedArray):
            mask = ~hidden_state.mask[0,:] # NOTE THE INVERTER 
            inds = np.nonzero([ mask[k]*mask[k+1] for k in range(len(mask)-1)])[0]
    
            X = np.mat(hidden_state[:,mask])
            n_pts = len(np.nonzero(mask)[0])
    
            Y = np.mat(obs[:,mask])
            if include_offset:
                X = np.vstack([ X, np.ones([1,n_pts]) ])
        else:
            num_hidden_state, n_pts = hidden_state.shape
            X = np.mat(hidden_state)
            if include_offset:
                X = np.vstack([ X, np.ones([1,n_pts]) ])
            Y = np.mat(obs)
        X = np.mat(X, dtype=np.float64)

        R = (X * X.T)
        S = (Y * X.T)
        T = (Y * Y.T)
        ESS = n_pts

        return (R, S, T, ESS)

    def init(self, decoder):
        '''
        Retrieve sufficient statistics from the seed decoder.

        Parameters
        ----------
        decoder : bmi.Decoder instance
            The seed decoder before any adaptation runs.

        Returns
        -------
        None
        '''
        self.R = decoder.filt.R
        self.S = decoder.filt.S
        self.T = decoder.filt.T
        self.ESS = decoder.filt.ESS


        #Neural indices that will be adapted / stable are defined here:
        self.feature_inds = np.arange(decoder.n_features)

        # Units that you want to stay stable
        self.stable_inds = []

        # By default, tuning parameters for all features will adapt
        if hasattr(decoder, 'adapting_neur_inds'):
            self.set_stable_inds(None, adapting_inds=decoder.adapting_neur_inds)
        else:
            self.adapting_inds = self.feature_inds.copy()
            self.stable_inds_independent = False

        self.adapting_inds_mesh = np.ix_(self.adapting_inds, self.adapting_inds)

        #Are stable units independent from other units ? If yes Q[stable_unit, other_units] = 0
        #State space indices that will be adapted: 
        self.state_inds = np.arange(len(decoder.states))
        if hasattr(decoder, 'adapting_state_inds'):
            if type(decoder.adapting_state_inds) is not list:
                ad = [i for i, j in enumerate(decoder.states) if j in decoder.adapting_state_inds.state_names]
            else:
                ad = decoder.adapting_state_inds
            self.set_stable_states(None, adapting_state_inds=ad)
        else:
            self.state_adapting_inds = np.arange(decoder.n_states)
        
        self.neur_by_state_adapting_inds_mesh = np.ix_(self.adapting_inds, self.state_adapting_inds)


        if hasattr(decoder, 'adapt_mFR_stats'):
            print('setitng adapting mFR. updater', decoder.adapt_mFR_stats)
            self.adapt_mFR_stats = decoder.adapt_mFR_stats
        else:
            self.adapt_mFR_stats = False

    def calc(self, intended_kin=None, spike_counts=None, decoder=None, half_life=None, values=None, **kwargs):
        '''
        Parameters
        ----------
        intended_kin : np.ndarray of shape (n_states, batch_size)
            Batch of estimates of intended kinematics, from the learner
        spike_counts : np.ndarray of shape (n_features, batch_size)
            Batch of observations of decoder features, from the learner
        decoder : bmi.Decoder instance
            Reference to the Decoder instance
        half_life : float, optional
            Half-life to use to calculate the parameter change step size [s]. If not specified, the half-life specified when the Updater was constructed is used.
        values : np.ndarray, optional
            Relative value of each sample of the batch. If not specified, each sample is assumed to have equal value.
        kwargs : dict
            Optional keyword arguments, ignored

        Returns
        -------
        new_params : dict
            New parameters to feed back to the Decoder in use by the task.
        '''
        if intended_kin is None or spike_counts is None or decoder is None:
            raise ValueError("must specify intended_kin, spike_counts and decoder objects for the updater to work!")

        # Calculate the step size based on the half life and the number of samples to train from
        batch_size = intended_kin.shape[1]
        batch_time = batch_size * decoder.binlen            

        if half_life is not None:
            rho = np.exp(np.log(0.5)/(half_life/batch_time))
        else:
            rho = self.rho 

        #update driver of neurons
        try:
            drives_neurons = decoder.drives_neurons.copy()
            mFR_old        = decoder.mFR.copy()
            sdFR_old       = decoder.sdFR.copy()
        except:
            drives_neurons = decoder.drives_neurons
            mFR_old = decoder.mFR
            sdFR_old = decoder.sdFR

        x = np.mat(intended_kin)
        y = np.mat(spike_counts)
        #limit x to the indices that can adapt:
        #x = x[self.state_adapting_inds, :]

        # limit y to the features which are permitted to adapt
        #y = y[self.adapting_inds, :]

        if values is not None:
            n_samples = np.sum(values)
            B = np.mat(np.diag(values))
        else:
            n_samples = spike_counts.shape[1]
            B = np.mat(np.eye(n_samples))

        if self.adapt_C_xpose_Q_inv_C:
            #self.R[self.state_adapting_inds_mesh] = rho*self.R[self.state_adapting_inds_mesh] + (x*B*x.T)
            self.R = rho*self.R + (x*B*x.T)

        if np.any(np.isnan(self.R)):
            print('np.nan in self.R in riglib/bmi/clda.py!')

        #self.S[self.neur_by_state_adapting_inds_mesh] = rho*self.S[self.neur_by_state_adapting_inds_mesh] + (y*B*x.T)
        #self.T[self.adapting_inds_mesh] = rho*self.T[self.adapting_inds_mesh] + np.dot(y, B*y.T)

        self.S[:, decoder.drives_neurons] = rho*self.S[:, decoder.drives_neurons] + (y*B*x[decoder.drives_neurons, :].T)
        self.T = rho*self.T + np.dot(y, B*y.T)
        self.ESS = rho*self.ESS + n_samples

        R_inv = np.mat(np.zeros(self.R.shape))
        
        try:
            if self.regularizer is None:
                R_inv[np.ix_(drives_neurons, drives_neurons)] = np.linalg.pinv(self.R[np.ix_(drives_neurons, drives_neurons)])
            else:
                dn = np.sum(drives_neurons)
                R_inv[np.ix_(drives_neurons, drives_neurons)] = np.linalg.pinv(self.R[np.ix_(drives_neurons, drives_neurons)]+self.regularizer*np.eye(dn))
        except:
            print(self.R)
            print('Error with pinv in riglib/bmi/clda.py')

        C_new = self.S * R_inv
        C = copy.deepcopy(decoder.filt.C)
        C[np.ix_(self.adapting_inds, self.state_adapting_inds)] = C_new[np.ix_(self.adapting_inds, self.state_adapting_inds)]
        
        Q = (1./self.ESS) * (self.T - self.S*C.T)
        if hasattr(self, 'stable_inds_mesh'):
            if len(self.stable_inds) > 0:
                print('stable inds mesh: ', self.stable_inds, self.stable_inds_mesh)
                Q_old = decoder.filt.Q[self.stable_inds_mesh].copy()
                Q[self.stable_inds_mesh] = Q_old

        if self.stable_inds_independent:
            Q[np.ix_(self.stable_inds, self.adapting_inds)] = 0
            Q[np.ix_(self.adapting_inds, self.stable_inds)] = 0

        #mFR and sdFR are not exempt from the 'adapting_inds'
        try:
            mFR = mFR_old.copy()
            sdFR = sdFR_old.copy()
        except:
            mFR = 0.
            sdFR = 1.

        if self.adapt_mFR_stats:
            mFR[self.adapting_inds] = (1-rho)*np.mean(spike_counts[self.adapting_inds,:].T, axis=0) + rho*mFR_old[self.adapting_inds]
            sdFR[self.adapting_inds] = (1-rho)*np.std(spike_counts[self.adapting_inds,:].T, axis=0) + rho*sdFR_old[self.adapting_inds]

        C_xpose_Q_inv = C.T * np.linalg.pinv(Q)
        new_params = {'filt.C':C, 'filt.Q':Q, 'filt.C_xpose_Q_inv':C_xpose_Q_inv,
            'mFR':mFR, 'sdFR':sdFR, 'kf.ESS':self.ESS, 'filt.S':self.S, 'filt.T':self.T}

        if self.adapt_C_xpose_Q_inv_C:
            C_xpose_Q_inv_C = C_xpose_Q_inv * C
            new_params['filt.C_xpose_Q_inv_C'] = C_xpose_Q_inv_C
            new_params['filt.C_xpose_Q_inv'] = C_xpose_Q_inv
            new_params['filt.R'] = self.R
        else:
            new_params['filt.C_xpose_Q_inv_C'] = decoder.filt.C_xpose_Q_inv_C
            new_params['filt.R'] = decoder.filt.R

        self._new_params = new_params
        return new_params

    def set_stable_inds(self, stable_inds, adapting_inds=None, stable_inds_independent=False):
        '''
        Set certain neural tuning parmeters to remain static, e.g., if you 
        want to add a new unit to a decoder but keep the existing parameters for the old units. 
        '''
        if adapting_inds is None: # Stable inds provided
            self.stable_inds = stable_inds   
            self.adapting_inds = np.array([x for x in self.feature_inds if x not in self.stable_inds]).astype(int)
        elif stable_inds is None: # Adapting inds provided:
            self.adapting_inds = np.array(adapting_inds).astype(int)
            self.stable_inds = np.array([x for x in self.feature_inds if x not in self.adapting_inds])

        self.adapting_inds_mesh = np.ix_(self.adapting_inds, self.adapting_inds)
        self.stable_inds_mesh = np.ix_(self.stable_inds, self.stable_inds)
        self.stable_inds_independent = stable_inds_independent

    def set_stable_states(self, stable_state_inds, adapting_state_inds=None, stable_state_inds_independent=False):
        '''
        Maybe you want to keep specific states states (e.g. in iBMI, keep ArmAssist stable but adapt ReHand)
        '''
        if adapting_state_inds is None:
            self.state_adapting_inds = np.array([x for x in self.state_inds if x not in stable_state_inds])
        elif stable_state_inds is None:
            self.state_adapting_inds = np.array(adapting_state_inds).astype(int)
        self.state_adapting_inds_mesh = np.ix_(self.state_adapting_inds, self.state_adapting_inds)
        self.stable_state_inds_independent = stable_state_inds_independent

class KFRML_IVC(KFRML):
    '''
    RML version where diagonality constraints are imposed on the steady state KF matrices
    '''
    default_gain = None
    def calc(self, intended_kin=None, spike_counts=None, decoder=None, half_life=None, values=None, **kwargs):
        '''
        See KFRML.calc for input argument documentation
        '''
        new_params = super(KFRML_IVC, self).calc(intended_kin=intended_kin, spike_counts=spike_counts, decoder=decoder, half_life=half_life, values=values, **kwargs)
        C, Q, = new_params['filt.C'], new_params['filt.Q']

        D = (C.T * np.linalg.pinv(Q) * C)
        if self.default_gain == None:
            # assume velocity states are last half of states: 
            v0 = int(.5*(D.shape[0] - 1))
            
            # get non-zero indices (e.g. cursor state only uses indices 3 and 5, not 4)
            vix = np.nonzero(np.diag(D[v0:-1, v0:-1]))[0] + v0

            # take mean: 
            d = np.mean(np.diag(D)[vix])

            # set diagonal to mean, off-diagonal to zeros: 
            D[v0:-1, v0:-1] = np.diag(np.zeros((v0,))+d)
            
            #Old: cursor only: 
            #d = np.mean([D[3,3], D[5,5]])
            #D[3:6, 3:6] = np.diag([d, d, d])
        else:
            # calculate the gain from the riccati equation solution
            A_diag = np.diag(np.asarray(decoder.filt.A[3:6, 3:6]))
            W_diag = np.diag(np.asarray(decoder.filt.W[3:6, 3:6]))
            D_diag = []
            for a, w, n in zip(A_diag, W_diag, [self.default_gain]*3):
                d = self.scalar_riccati_eq_soln(a, w, n)
                D_diag.append(d)

            D[3:6, 3:6] = np.mat(np.diag(D_diag))

        new_params['filt.C_xpose_Q_inv_C'] = D
        new_params['filt.C_xpose_Q_inv'] = C.T * np.linalg.pinv(Q)
        return new_params

    @classmethod
    def scalar_riccati_eq_soln(cls, a, w, n):
        '''
        For the scalar case, determine what you want the prediction covariance of the KF, 
        which follows the riccati recursion for constant model parameters,
        based on what gain you want to set for the steady-state KF

        Parameters
        ----------
        a : float
            Diagonal value of the A matrix for the velocity terms
        w : float
            Diagonal value of the W matrix for the velocity terms
        n : float
            Steady-state kalman filter gain for the velocity terms

        Returns 
        -------
        float
        '''
        return (1-a*n)/w * (a-n)/n         

class KFRML_baseline(KFRML):
    '''
    RML version where only the baseline firing rates are adapted
    '''
    def calc(self, intended_kin=None, spike_counts=None, decoder=None, half_life=None, values=None, **kwargs):
        '''
        See KFRML.calc for input argument documentation
        '''
        print("calculating new baseline parameters")
        if half_life is not None:
            rho = np.exp(np.log(0.5)/(half_life/self.batch_time))
        else:
            rho = self.rho 

        drives_neurons = decoder.drives_neurons
        mFR_old        = decoder.mFR.copy()
        sdFR_old       = decoder.sdFR.copy()
        
        mFR = mFR_old.copy()
        sdFR= sdFR_old.copy()

        mFR[self.adapting_inds] = (1-rho)*np.mean(spike_counts[self.adapting_inds,:].T, axis=0) + rho*mFR_old[self.adapting_inds]
        sdFR[self.adapting_inds] = (1-rho)*np.std(spike_counts[self.adapting_inds,:].T, axis=0) + rho*sdFR_old[self.adapting_inds]
        
        new_params = {'mFR':mFR, 'sdFR':sdFR}

        return new_params


###################################
##### Updaters in development #####
###################################
class PPFRML(Updater):
    '''RML method applied to more generic GLM'''
    update_kwargs = dict()
    def __init__(self, *args, **kwargs):
        super(PPFRML, self).__init__(self.calc, multiproc=False)

    def init(self, decoder):
        self.dt = decoder.filt.dt
        self.C_est = decoder.filt.C
        self.H = decoder.H
        self.M = decoder.M
        self.S = decoder.S
        
        self.neuron_driving_state_inds = np.nonzero(decoder.drives_neurons)[0]
        self.neuron_driving_states = list(np.take(decoder.states, np.nonzero(decoder.drives_neurons)[0]))
        self.n_states = len(decoder.states)
        self.full_size = len(decoder.states)
        

    def calc(self, intended_kin=None, spike_counts=None, decoder=None, half_life=120., **kwargs):
        '''
        # time iterative RLS
        '''
        if (intended_kin is None) or (spike_counts is None) or (decoder is None):
            raise ValueError("must specify intended_kin, spike_counts and decoder objects for the updater to work!")        

        batch_size = 1.
        batch_time = batch_size * decoder.binlen    
        rho = np.exp(np.log(0.5)/(half_life/batch_time))
    
        n_cells = self.C_est.shape[0]
        n_obs = intended_kin.shape[1]
        intended_kin = np.mat(intended_kin)
        # print "updating"
        # print intended_kin
        # print spike_counts.T
        spike_counts[spike_counts > 1] = 1
        for k in range(n_cells):
            for m in range(n_obs):
                H = np.mat(self.H[k])
                S = np.mat(self.S[k].reshape(-1,1))
                M = np.mat(self.M[k].reshape(-1,1))
                c = self.C_est[k, self.neuron_driving_state_inds].T

                # print H
                c_new = c - H.I * (S - M)
                c = rho*c + (1-rho)*c_new
                self.C_est[k, self.neuron_driving_state_inds] = c.T
                
                x_m = intended_kin[self.neuron_driving_state_inds, m] #X[k].T
                mu_m = np.exp(c.T * x_m)[0,0]
                y_m = spike_counts[k, m]


                self.H[k] = rho*H + (1-rho)*(-mu_m * x_m * x_m.T)
                self.M[k] = np.array(rho*M + (1-rho)*(mu_m * x_m)).ravel()
                self.S[k] = np.array(rho*S + (1-rho)*(y_m*x_m)).ravel()

        return {'filt.C': self.C_est}


###############################
##### Deprecated updaters #####
###############################
class KFSmoothbatch(Updater):
    '''
    Deprecation Warning: This update method has not been used for quite long. See KFRML for an enhanced but similar method

    Calculate KF Parameter updates using the SmoothBatch method. See [Orsborn et al, 2012] for mathematical details
    '''
    update_kwargs = dict(steady_state=True)
    def __init__(self, batch_time, half_life):
        '''
        Constructor for KFSmoothbatch

        Parameters
        ----------
        batch_time : float
            Time over which to collect sample data
        half_life : float
            Time over which parameters are half-overwritten

        Return
        ------
        KFSmoothbatch instance
        '''
        super(KFSmoothbatch, self).__init__(self.calc, multiproc=False)
        self.half_life = half_life
        self.batch_time = batch_time
        self.rho = np.exp(np.log(0.5) / (self.half_life/batch_time))
        
    def calc(self, intended_kin=None, spike_counts=None, decoder=None, half_life=None, **kwargs):
        """
        Smoothbatch calculations

        Run least-squares on (intended_kinematics, spike_counts) to 
        determine the C_hat and Q_hat of new batch. Then combine with 
        old parameters using step-size rho
        """
        print("calculating new SB parameters")
        C_old          = decoder.kf.C
        Q_old          = decoder.kf.Q
        drives_neurons = decoder.drives_neurons
        mFR_old        = decoder.mFR
        sdFR_old       = decoder.sdFR

        C_hat, Q_hat = kfdecoder.KalmanFilter.MLE_obs_model(
            intended_kin, spike_counts, include_offset=False, drives_obs=drives_neurons)

        if not (half_life is None):
            rho = np.exp(np.log(0.5)/(half_life/self.batch_time))
        else:
            rho = self.rho 

        C = (1-rho)*C_hat + rho*C_old
        Q = (1-rho)*Q_hat + rho*Q_old

        mFR = (1-rho)*np.mean(spike_counts.T, axis=0) + rho*mFR_old
        sdFR = (1-rho)*np.std(spike_counts.T, axis=0) + rho*sdFR_old
        
        D = C.T * np.linalg.pinv(Q) * C
        new_params = {'kf.C':C, 'kf.Q':Q, 
            'kf.C_xpose_Q_inv_C':D, 'kf.C_xpose_Q_inv':C.T * np.linalg.pinv(Q),
            'mFR':mFR, 'sdFR':sdFR, 'rho':rho }
        return new_params


class KFOrthogonalPlantSmoothbatch(KFSmoothbatch):
    '''This module is deprecated. See KFRML_IVC'''
    def __init__(self, *args, **kwargs):
        self.default_gain = kwargs.pop('default_gain', None)
        suoer(KFOrthogonalPlantSmoothbatch, self).__init__(*args, **kwargs)

    def calc(self, *args, **kwargs):
        new_params = super(KFOrthogonalPlantSmoothbatch, self).calc(*args, **kwargs)
        C, Q, = new_params['kf.C'], new_params['kf.Q']

        D = (C.T * np.linalg.pinv(Q) * C)
        if self.default_gain == None:
            d = np.mean([D[3,3], D[5,5]])
            D[3:6, 3:6] = np.diag([d, d, d])
        else:
            # calculate the gain from the riccati equation solution
            A_diag = np.diag(np.asarray(decoder.filt.A[3:6, 3:6]))
            W_diag = np.diag(np.asarray(decoder.filt.W[3:6, 3:6]))
            D_diag = []
            for a, w, n in zip(A_diag, W_diag, self.default_gain):
                d = self.scalar_riccati_eq_soln(a, w, n)
                D_diag.append(d)

            D[3:6, 3:6] = np.mat(np.diag(D_diag))

        new_params['kf.C_xpose_Q_inv_C'] = D
        new_params['kf.C_xpose_Q_inv'] = C.T * np.linalg.pinv(Q)
        return new_params


class PPFSmoothbatch(Updater):
    '''
    Deprecated: This updater as of 2015-Sept-19 was never used in an experiment. 
    '''
    def __init__(self, batch_time, half_life):
        super(PPFSmoothbatch, self).__init__(self.calc, multiproc=True)
        self.half_life = half_life
        self.rho = np.exp(np.log(0.5) / (self.half_life/batch_time))

    def calc(self, intended_kin=None, spike_counts=None, decoder=None, half_life=None, **kwargs):
        """
        Smoothbatch calculations

        Run least-squares on (intended_kinematics, spike_counts) to 
        determine the C_hat and Q_hat of new batch. Then combine with 
        old parameters using step-size rho
        """
        if half_life is not None:
            rho = np.exp(np.log(0.5)/(half_life/self.batch_time))
        else:
            rho = self.rho 

        C_old = decoder.filt.C
        drives_neurons = decoder.drives_neurons
        states = decoder.states
        decoding_states = np.take(states, np.nonzero(drives_neurons)).ravel().tolist() #['hand_vx', 'hand_vz', 'offset'] 

        C_hat, pvalues = ppfdecoder.PointProcessFilter.MLE_obs_model(
            intended_kin, spike_counts, include_offset=False, drives_obs=drives_neurons)
        C_hat = train.inflate(C_hat, decoding_states, states, axis=1)
        pvalues = train.inflate(pvalues, decoding_states, states, axis=1)
        pvalues[pvalues[:,:-1] == 0] = np.inf

        mesh = np.nonzero(pvalues < 0.1)
        C = np.array(C_old.copy())
        C[mesh] = (1-rho)*C_hat[mesh] + rho*np.array(C_old)[mesh]
        C = np.mat(C)

        new_params = {'filt.C':C}
        return new_params




import tables
import numpy as np
import tempfile
import datetime
import os

compfilt = tables.Filters(complevel=5, complib="zlib", shuffle=True)

class SupplementaryHDF(object):

    def __init__(self, channels, sink_dtype, source, data_dir='/var/tmp/'):

        dt = datetime.datetime.now()
        tm = dt.time()
        subject = 'test' if not hasattr(source, 'subj') else source.subj
        saveid = 'None' if not hasattr(source, 'saveid') else source.saveid
        self.filename = f'tmp_{str(source)}_{subject}_{saveid}.hdf'
        self.h5_file = tables.open_file(os.path.join(data_dir, self.filename), "w")

        #If sink datatype is not specified: 
        if sink_dtype is None:
            self.dtype = np.dtype([('data',       np.float64),
                              ('ts_arrival', np.float64)])

            self.send_to_sinks_dtype = np.dtype([('chan'+str(chan), self.dtype) for chan in channels])

        else:
            self.send_to_sinks_dtype = sink_dtype
        self.supp_data = self.h5_file.create_table("/", "data", self.send_to_sinks_dtype, filters=compfilt)
        self.supp_data.attrs['samplerate'] = source.update_freq
        self.supp_data.attrs['channels'] = channels
        self.supp_data.attrs['dtype'] = source.dtype
        
        if hasattr(source, 'n_arrays'):
            self.supp_data.attrs['n_arrays'] = source.n_arrays

    def add_data(self, data):
        self.supp_data.append(data)

    def close_data(self):
        self.h5_file.close()
        print("Closed supplementary hdf file")
